<?php

namespace Dog;

use Dog\Exception\RoyalException;
use Dog\Helper\Account;
use Dog\Helper\Teams;
use Dog\Utils\UI;
use FML\Controls\Frame;
use FML\Controls\FrameInstance;
use FML\Controls\Label;
use FML\Controls\Quad;
use FML\Elements\FrameModel;
use FML\ManiaLink;
use FML\Script\ScriptFunction;
use FML\Script\ScriptInclude;
use FML\Script\ScriptLabel;
use ManiaControl\Admin\AuthenticationManager;
use ManiaControl\Callbacks\CallbackListener;
use ManiaControl\Callbacks\CallbackManager;
use ManiaControl\Callbacks\Callbacks;
use ManiaControl\Callbacks\Structures\TrackMania\OnScoresStructure;
use ManiaControl\Callbacks\TimerListener;
use ManiaControl\Commands\CommandListener;
use ManiaControl\Files\AsyncHttpRequest;
use ManiaControl\Logger;
use ManiaControl\ManiaControl;
use ManiaControl\Manialinks\ManialinkPageAnswerListener;
use ManiaControl\Maps\MapCommands;
use ManiaControl\Players\Player;
use ManiaControl\Players\PlayerManager;
use ManiaControl\Plugins\Plugin;

/**
 * TM2020 Royal mode support
 *
 * @author Dog
 * @version 0.51
 */
class Royal implements Plugin, CallbackListener, ManialinkPageAnswerListener, CommandListener, TimerListener
{
	const ID = 180;
	const VERSION = 0.51;
	const NAME = 'Royal';
	const AUTHOR = 'Dog';

	const MLID_TEAMDISPLAY = 'Royal.TeamDisplay';
	const MLID_ADMIN_UI = 'Royal.Admin';

	const ML_ACTION_STARTMATCH = 'Royal.Admin.StartMatch';
	const ML_ACTION_RESETTEAMS = 'Royal.Admin.ResetTeams';
	const ML_ACTION_RANDOMIZETEAMS = 'Royal.Admin.RandomizeTeams';
	const ML_ACTION_FORCEPLAYERINTOTEAM = 'Royal.Admin.ForcePlayerIntoTeam';
	const ML_ACTION_REMOVEPLAYERFROMTEAM = 'Royal.Admin.RemovePlayerFromTeam';

	const ML_ACTION_CLOSE_TEAMDISPLAY = 'Royal.Player.CloseTeamDisplay';
	const ML_ACTION_JOINTEAM = 'Royal.Player.JoinTeam';
	const ML_ACTION_LEAVETEAM = 'Royal.Player.LeaveTeam';

	const XMLRPC_METHOD_ADDPLAYER = 'Club.Match.AddPlayer';
	const XMLRPC_METHOD_REMOVEPLAYER = 'Club.Match.RemovePlayer';

	const CB_PLAYER_FORCED = 'Royal.ForcePlayer';

	const SETTING_TEAMDISPLAY_COLS = 'Number of columns';
	const SETTING_TEAMDISPLAY_POSX = 'X Position of the team display';
	const SETTING_TEAMDISPLAY_POSY = 'Y Position of the team display';

	const SETTING_POST_ENDPOINT = 'POST endpoint url';
	const SETTING_POST_ENDPOINT_HEADER_KEY = 'POST endpoint header key';
	const SETTING_POST_ENDPOINT_HEADER_VALUE = 'POST endpoint header value';

	const SETTING_LOADINGSCREEN_IMAGE_URLS = 'Comma separated list of image urls';

	//Match status
	const MATCH_INPROGRESS = 'Royal.MatchInProgress';
	const MATCH_ENDED = 'Royal.MatchEnded';
	const MATCH_INVALID = 'Royal.MatchInvalid';

	/**
	 * Private Properties
	 */
	/** @var ManiaControl $maniaControl */
	private $maniaControl = null;

	private $matchStatus;

	private bool $isRoyalMode;

	/**
	 * @see \ManiaControl\Plugins\Plugin::prepare()
	 */
	public static function prepare(ManiaControl $maniaControl)
	{
	}

	/**
	 * @see \ManiaControl\Plugins\Plugin::load()
	 */
	public function load(ManiaControl $maniaControl)
	{
		$this->maniaControl = $maniaControl;

		// Player Commands
		$this->maniaControl->getCommandManager()->registerCommandListener(['teamselect', 'selectteam', 'display'], $this, 'onCommandShowTeamSelection', false, 'Display team selection window');
		// Admin Commands
		$this->maniaControl->getCommandManager()->registerCommandListener(['start', 'startroyal', 'royalstart'], $this, 'onCommandStart', true, 'Start the game');
		$this->maniaControl->getCommandManager()->registerCommandListener(['displayall', 'teamselectall'], $this, 'onCommandShowTeamSelectionAll', true, 'Show team selection window to all');
		$this->maniaControl->getCommandManager()->registerCommandListener('addplayer', $this, 'onCommandAddPlayer', true);
		// ManiaControl Commands
		$this->maniaControl->getCommandManager()->registerCommandListener(['restartmap', 'resmap', 'res'], $this, 'onCommandRestart', true);
		$this->maniaControl->getCommandManager()->registerCommandListener(['nextmap', 'next', 'skip'], $this, 'onCommandRestart', true);

		// Callbacks
		$this->maniaControl->getCallbackManager()->registerCallbackListener(CallbackManager::CB_MP_PLAYERMANIALINKPAGEANSWER, $this, 'handleManialinkPageAnswer');
		$this->maniaControl->getCallbackManager()->registerCallbackListener(PlayerManager::CB_PLAYERCONNECT, $this, 'handlePlayerConnect');
		$this->maniaControl->getCallbackManager()->registerCallbackListener(PlayerManager::CB_PLAYERDISCONNECT, $this, 'handlePlayerDisconnect');
		$this->maniaControl->getCallbackManager()->registerCallbackListener(Callbacks::TM_SCORES, $this, 'handleScores');

		$this->maniaControl->getCallbackManager()->registerScriptCallbackListener("Maniaplanet.StartMap_Start", $this, 'handleMapStart');
		// $this->maniaControl->getCallbackManager()->registerCallbackListener(Callbacks::SCRIPTCALLBACK, $this, 'callbacks');


		$this->maniaControl->getManialinkManager()->registerManialinkPageAnswerListener(MapCommands::ACTION_RESTART_MAP, $this, 'onCommandRestart');
		$this->maniaControl->getManialinkManager()->registerManialinkPageAnswerListener(MapCommands::ACTION_SKIP_MAP, $this, 'onCommandRestart');

		// Plugin related callbacks
		$this->maniaControl->getCallbackManager()->registerCallbackListener(self::CB_PLAYER_FORCED, $this, 'handleForcePlayer');

		// Timers
		$this->maniaControl->getTimerManager()->registerTimerListening($this, 'handle60Secs', 60000);

		// Settings
		$this->maniaControl->getSettingManager()->initSetting($this, self::SETTING_TEAMDISPLAY_COLS, 2, '');
		$this->maniaControl->getSettingManager()->initSetting($this, self::SETTING_TEAMDISPLAY_POSX, 40, '');
		$this->maniaControl->getSettingManager()->initSetting($this, self::SETTING_TEAMDISPLAY_POSY, -20, '');
		$this->maniaControl->getSettingManager()->initSetting($this, self::SETTING_POST_ENDPOINT, '', '');
		$this->maniaControl->getSettingManager()->initSetting($this, self::SETTING_POST_ENDPOINT_HEADER_KEY, '', '');
		$this->maniaControl->getSettingManager()->initSetting($this, self::SETTING_POST_ENDPOINT_HEADER_VALUE, '', '');
		$this->maniaControl->getSettingManager()->initSetting($this, self::SETTING_LOADINGSCREEN_IMAGE_URLS, '', '');

		Teams::Init();
		$this->updateTeamSelection();

	}

	/**
	 * @see \ManiaControl\Plugins\Plugin::unload()
	 */
	public function unload()
	{
		$this->maniaControl->getManialinkManager()->hideManialink(self::MLID_TEAMDISPLAY);
		$this->maniaControl->getManialinkManager()->hideManialink(self::MLID_TEAMDISPLAY . '.Update');

		Teams::Init();
	}

	/**
	 * @see \ManiaControl\Plugins\Plugin::getId()
	 */
	public static function getId()
	{
		return self::ID;
	}

	/**
	 * @see \ManiaControl\Plugins\Plugin::getName()
	 */
	public static function getName()
	{
		return self::NAME;
	}

	/**
	 * @see \ManiaControl\Plugins\Plugin::getVersion()
	 */
	public static function getVersion()
	{
		return self::VERSION;
	}

	/**
	 * @see \ManiaControl\Plugins\Plugin::getAuthor()
	 */
	public static function getAuthor()
	{
		return self::AUTHOR;
	}

	/**
	 * @see \ManiaControl\Plugins\Plugin::getDescription()
	 */
	public static function getDescription()
	{
		return 'TM2020 Royal Mode support.';
	}

/* 	public function callbacks($callback)
	{
		var_dump($callback);
	} */

	public function handleMapStart($callback)
	{
		//var_dump($this->isRoyalMode, $this->matchStatus);
		if ($this->isRoyalMode && $this->matchStatus != self::MATCH_INPROGRESS) {
			$players = $this->maniaControl->getPlayerManager()->getPlayers();
			foreach ($players as $player) {
				$this->teamSelection($player);
			}
		}

		$images = str_getcsv(preg_replace("/\s*,\s*/", ",", $this->maniaControl->getSettingManager()->getSettingValue($this, self::SETTING_LOADINGSCREEN_IMAGE_URLS)));

		// no image urls set
		if (in_array(null, $images)) {
			return;
		}

		// grab a random image
		$key = array_rand($images);
		$image = $images[$key];

		Logger::log("Setting image url to: " . $image);

		$this->maniaControl->getClient()->setModeScriptSettings([
			'S_LoadingScreenImageUrl' => $image
		]);
	}

	// Fetch the script name every minute
	public function handle60Secs()
	{
		$script = $this->maniaControl->getClient()->getScriptName();
		$scriptName = str_replace(".Script.txt", "", $script["CurrentValue"]);
		// Check if the current mode is royal (be it custom or official)
		if (strpos($scriptName, "TM_Royal_") !== false) {
			return $this->isRoyalMode = true;
		}
		return $this->isRoyalMode = false;
	}

	public function handleForcePlayer($callback, Player $caller)
	{
		//var_dump($callback);
		$team = $callback["team"];
		$accountIds = $callback["accountIds"];

		$accounts = explode(",", $accountIds);

		foreach ($accounts as $accountId) {
			$player = $this->maniaControl->getPlayerManager()->getPlayer(Account::toLogin($accountId));
			if (!is_null($player)) {
				try {
					Teams::AddPlayer($player, ucwords($team, "_"));
				} catch (\Exception $e) {
					Teams::RemovePlayer($player);
					Teams::AddPlayer($player, $team);
				}
			} else {
				try {
					$this->maniaControl->getClient()->triggerModeScriptEvent(self::XMLRPC_METHOD_ADDPLAYER, [
						$accountId,
						(string) Teams::GetTeamIdV2($team)
					], true);
				} catch (\Exception $e) {
					Logger::logError($e->getMessage());
				}
			}
		}

		//echo "[HandleForcePlayer] received team Id $team for accounts: $accountIds" . PHP_EOL;
	}

	public function handleScores(OnScoresStructure $struct)
	{
		// The Scores Structure is used to determine the match end.
		// If a POST url is set, the plugin will attempt to POST
		// scores to the specified url.
		if ($struct->getSection() != "EndMatchEarly")
			return;
		$scores = $struct->getPlainJsonObject();
		// Do stuff with $scores
		$url = $this->maniaControl->getSettingManager()->getSettingValue($this, self::SETTING_POST_ENDPOINT);
		if ($url != "" && $this->matchStatus != self::MATCH_INVALID) {
			$this->sendPostRequest(
				$scores,
				$url
			);
		}
		// Set match status
		$this->matchStatus = self::MATCH_ENDED;
		$this->maniaControl->getChat()->sendInformation("Match has ended");
	}

	public function handlePlayerConnect(Player $player)
	{
		$this->updateTeamSelection();
		if ($this->isRoyalMode && $this->matchStatus == self::MATCH_ENDED) {
			$this->teamSelection($player);
		}
	}

	public function handlePlayerDisconnect(Player $player)
	{
		Teams::RemovePlayer($player);
		$this->updateTeamSelection();
	}

	public function handleManialinkPageAnswer($callback)
	{
		//var_dump($callback);

		$login = $callback[1][1];
		$action = $callback[1][2];
		$actionArray = explode('.', $action);

		if (count($actionArray) > 3) {
			$team = $actionArray[3];
			unset($actionArray[count($actionArray) - 1]);
			$action = implode('.', $actionArray);
		}

		$player = $this->maniaControl->getPlayerManager()->getPlayer($login);

		switch ($action) {
			case self::ML_ACTION_JOINTEAM:
				if ($this->matchStatus == self::MATCH_INPROGRESS) {
					// The match is in progress, players shouldn't be able to join/switch/leave teams.
					$this->maniaControl->getChat()->sendInformation("Cannot join a team while a match is in progress.", $player->login);
				} else {
					try {
						Teams::AddPlayer($player, $team);
					} catch (RoyalException $e) {
						// Exception thrown when the player is already part of a team.
						// To keep things dynamic, allow players to switch without having to leave first.
						Teams::RemovePlayer($player);
						Teams::AddPlayer($player, $team);
					}
					$this->updateTeamSelection();
				}
				break;
			case self::ML_ACTION_LEAVETEAM:
				if ($this->matchStatus == self::MATCH_INPROGRESS) {
					// The match is in progress, players shouldn't be able to join/switch/leave teams.
					$this->maniaControl->getChat()->sendInformation("Cannot leave a team while a match is in progress.", $player->login);
				} else {
					Teams::RemovePlayer($player);
					$this->updateTeamSelection();
				}
				break;
			case self::ML_ACTION_STARTMATCH:
				$this->startGame($player);
				break;
			case self::ML_ACTION_RESETTEAMS:
				Teams::Init();
				$this->updateTeamSelection();
				break;
			case self::ML_ACTION_CLOSE_TEAMDISPLAY:
				$this->maniaControl->getManialinkManager()->hideManialink(self::MLID_TEAMDISPLAY . '.Update', $player->login);
				$this->maniaControl->getManialinkManager()->hideManialink(self::MLID_TEAMDISPLAY, $player->login);
				break;
		}
	}

	public function onCommandShowTeamSelectionAll($callback, Player $player)
	{
		if (!$this->maniaControl->getAuthenticationManager()->checkRight($player, AuthenticationManager::AUTH_LEVEL_ADMIN)) {
			$this->maniaControl->getAuthenticationManager()->sendNotAllowed($player);
			return;
		}

		$this->updateTeamSelection();
		foreach ($this->maniaControl->getPlayerManager()->getPlayers() as $player) {
			$this->teamSelection($player);
		}
	}

	public function onCommandShowTeamSelection($callback, Player $player)
	{
		$this->updateTeamSelection();
		$this->teamSelection($player);
	}

	public function onCommandRestart($callback, Player $player)
	{
		$this->matchStatus = self::MATCH_INVALID;
	}

	public function onCommandAddPlayer($callback, Player $player)
	{
		if (!$this->maniaControl->getAuthenticationManager()->checkRight($player, AuthenticationManager::AUTH_LEVEL_ADMIN)) {
			$this->maniaControl->getAuthenticationManager()->sendNotAllowed($player);
			return;
		}

		$command = $callback[1][2];
		// Clean up malformed csv
		$command = preg_replace("/\s*,\s*/", ",", $command);
		// Split the command into its individual parts
		$commandParts = explode(" ", $command);
		$team = $commandParts[1];
		$accountIds = $commandParts[2];

		if (!in_array(strtolower($team), array_map('strtolower', Teams::GetAvailableTeams()))) {
			return $this->maniaControl->getChat()->sendUsageInfo(
				sprintf("%s doesn't exist. Available Teams: %s", $team, implode(", ", Teams::GetAvailableTeams())),
				$player->login
			);
		}

		return $this->maniaControl->getCallbackManager()->triggerCallback(self::CB_PLAYER_FORCED, [
			"team" => $team,
			"accountIds" => $accountIds
		], $player);
	}

	public function onCommandStart($callback, Player $player)
	{
		$this->startGame($player);
	}

	public function updateTeamSelection()
	{
		$manialink = new ManiaLink(self::MLID_TEAMDISPLAY . '.Update');
		$script = $manialink->getScript();

		$script->addScriptFunction(ScriptFunction::class, "
			#Struct K_RoyalTeam {
				Text Members;
			}
		");
		$script->addScriptFunction(ScriptFunction::class, "
			#Struct K_ServerInfo {
				Integer NbPlayers;
				Integer NbPlayersInTeams;
			}
		");

		$script->addCustomScriptLabel(ScriptLabel::LOOP, '
			declare K_RoyalTeam[Text] RoyalTeams for This;
			declare K_ServerInfo ServerInfo for This;

			declare Text json = """' . Teams::toJSON() . '""";
			RoyalTeams.fromjson(json);

			ServerInfo = K_ServerInfo {
				NbPlayers = Players.count,
				NbPlayersInTeams = ' . count(array_merge(...array_values(Teams::Pending()))) . '
			};

		');

		$this->maniaControl->getManialinkManager()->sendManialink($manialink);
	}

	public function teamSelection(Player $player)
	{
		$manialink = new ManiaLink(self::MLID_TEAMDISPLAY, 3);

		$frameGlobal = new Frame("frame-global");
		$manialink->addChild($frameGlobal);
		$frameGlobal->setPosition(
			$this->maniaControl->getSettingManager()->getSettingValue($this, self::SETTING_TEAMDISPLAY_POSX),
			$this->maniaControl->getSettingManager()->getSettingValue($this, self::SETTING_TEAMDISPLAY_POSY)
		);

		$frameTeams = new Frame("frame-teams");
		$frameGlobal->addChild($frameTeams);

		$frameModel = new FrameModel("framemodel-team");
		$manialink->addChild($frameModel);

		// Close button
		$closeButton = UI::SquareButton("", self::ML_ACTION_CLOSE_TEAMDISPLAY, UI::ColorDarkRed, 8., 42.5, 80);
		$frameGlobal->addChild($closeButton);

		// Admin Buttons
		if ($this->maniaControl->getAuthenticationManager()->checkRight($player, AuthenticationManager::AUTH_LEVEL_ADMIN)) {
			$startButton = UI::Button("Start Game", self::ML_ACTION_STARTMATCH, UI::ColorDarkRed, -40, 93);
			$resetButton = UI::Button("Reset Teams", self::ML_ACTION_RESETTEAMS, UI::ColorGreenFour, -4, 93);
			// TODO: Implement randomize teams function
			//$randomizeButton = UI::Button("Randomize", self::ML_ACTION_RANDOMIZETEAMS, UI::ColorDarkBlue, 31.5, 93);

			$frameGlobal->addChildren([$startButton, $resetButton]);
		}
		// Server Info
		$frameInfo = new Frame("frame-info");
		$frameGlobal->addChild($frameInfo);

		// Info Background
		$infoBg = new Quad("info-background");
		$frameInfo->addChild($infoBg);
		$infoBg->setSize(165, 8)->setBackgroundColor("000514")
			->setOpacity(0.5)->setScriptEvents(true)->setAlign("center", "bottom")
			->setPosition(-44, 80);

		// Player stats info
		$infoLabel = new Label("info-label-player-stats");
		$frameInfo->addChild($infoLabel);
		$infoLabel->setSize(77, 8)->setTextSize(1)->setAlign("left", "center")
			->setTextPrefix('$i$t')->setTextFont("GameFontExtraBold")->setOpacity(1.)
			->setPosition(29, 84);

		// Info Text
		$infoText = clone $infoLabel;
		$frameInfo->addChild($infoText);
		$infoText->setId("info-label-text")->setPosition(-125, 84);

		/** Team FrameModel Start */
		// Background
		$teamBg = new Quad("team-background");
		$frameModel->addChild($teamBg);
		$teamBg->setSize(20, 8)->setSize(77, 8)->setBackgroundColor("000514")
			->setOpacity(0.5)->setScriptEvents(true)->setAlign("center", "bottom");

		// Logo Bg
		$logoBg = new Quad("logo-background");
		$frameModel->addChild($logoBg);
		$logoBg->setZ(1)->setSize(8, 8)->setX(42.5)
			->setBackgroundColor("000514")->setOpacity(0.6)->setAlign("center", "bottom");

		// Logo
		$logo = new Quad("logo");
		$frameModel->addChild($logo);
		$logo->setZ(1)->setSize(8, 8)->setPosition(42.5, 4)
			->setOpacity(1.)->setAlign("center", "center");

		// Members Label
		$teamLabel = new Label("team-label");
		$frameModel->addChild($teamLabel);
		$teamLabel->setSize(77, 8)->setTextSize(1)->setAlign("left", "center")
			->setTextPrefix('$i$t')->setTextFont("GameFontExtraBold")->setOpacity(1.)
			->setPosition(-37, 4);
		/** Team FrameModel End */

		// Leave Team Button
		$leaveButton = UI::Button("Leave Team", self::ML_ACTION_LEAVETEAM, UI::ColorDarkBlue, -40, -50);
		$frameGlobal->addChild($leaveButton);

		// Overlay
		$overlay = new Frame('overlay');
		$manialink->addChild($overlay);
		$overlay->setVisible(false)->setZ(5);

		// Overlay Background
		$overlayBg = new Quad('overlay-bg');
		$overlay->addChild($overlayBg);
		$overlayBg->setSize(85, 8)->setBackgroundColor("000514")->setAlign("center", "bottom");

		// OverlayBackground
		$overlayTri = new Quad("overlay-tri");
		$overlay->addChild($overlayTri);
		$overlayTri->setSize(3, 3)->setBackgroundColor("000514")->setAlign("center", "bottom")
			->setRotation(45.)->setY(-1.);

		// Overlay Member Names
		$overlayNames = new Label('overlay-label');
		$overlay->addChild($overlayNames);
		$overlayNames->setSize(83.9, 8)->setTextSize(1)->setAlign("left", "bottom")->setTextPrefix('$i$t')->setTextFont("GameFontExtraBold")->setAutoNewLine(true);

		$teams = Teams::Pending();
		$y = 50;
		$i = 0;
		$cols = $this->maniaControl->getSettingManager()->getSettingValue($this, self::SETTING_TEAMDISPLAY_COLS);
		if ($cols < 1)
			$cols = 2;

		foreach ($teams as $team => $val) {
			$x = -88 - ($i % $cols) * -88;
			if ($i % $cols > 0)
				$y += 10;

			$frameTeam = new Frame("frame-team-$team");
			$frameInstance = new FrameInstance("frame-instance-$team", "framemodel-team");
			$frameTeam->addChild($frameInstance);
			$frameTeams->addChild($frameTeam);

			$frameInstance->setPosition($x, $y);

			$y -= 10;
			$i++;
		}

		$script = $manialink->getScript();
		$script->setScriptInclude(ScriptInclude::TEXTLIB);

		$script->addScriptFunction(ScriptFunction::class, '
			#Struct K_RoyalTeam {
				Text Members;
			}
		');

		$script->addScriptFunction(ScriptFunction::class, '
			#Struct K_ServerInfo {
				Integer NbPlayers;
				Integer NbPlayersInTeams;
			}
		');

		$script->addScriptFunction(ScriptFunction::class, '
			Void UpdateState(CMlFrame _Frame_Team, Text _TeamName, Boolean _IsMouseOver, Boolean _IsMouseOut) {
				declare K_RoyalTeam[Text] RoyalTeams for This;
				declare CMlQuad Logo <=> (_Frame_Team.GetFirstChild("logo") as CMlQuad);
				declare CMlQuad Background <=> (_Frame_Team.GetFirstChild("team-background") as CMlQuad);
				declare CMlLabel Team_Label <=> (_Frame_Team.GetFirstChild("team-label") as CMlLabel);
				declare CMlFrame Overlay <=> (Page.GetFirstChild("overlay") as CMlFrame);
				declare CMlQuad Overlay_Background <=> (Page.GetFirstChild("overlay-bg") as CMlQuad);
				declare CMlQuad Overlay_Tri <=> (Page.GetFirstChild("overlay-tri") as CMlQuad);
				declare CMlLabel Overlay_Names <=> (Page.GetFirstChild("overlay-label") as CMlLabel);

				if (RoyalTeams.existskey(_TeamName)) {
					Logo.ImageUrl = """file://Media/Manialinks/Nadeo/ModeLibs/Clans/{{{_TeamName}}}.dds""";
					Team_Label.Value = ""^RoyalTeams[_TeamName].Members;

					if (Team_Label.ComputeWidth(Team_Label.Value) > 60.) {
						Team_Label.Value = TextLib::SubString(Team_Label.Value, 0, 40) ^ "...";
					}
				}

				if (_IsMouseOver) {
					AnimMgr.Add(Background, "<quad opacity=\'0.8\' />", 350, CAnimManager::EAnimManagerEasing::ExpOut);
					if (Team_Label.ComputeWidth(Team_Label.Value) > 60.) {
						AnimMgr.Add(Overlay, "<frame scale=\'1.\' />", 500, CAnimManager::EAnimManagerEasing::ExpOut);
						Overlay.RelativePosition_V3 = <Background.AbsolutePosition_V3.X + 4.,Background.AbsolutePosition_V3.Y + 10.>;
						Overlay.Visible = True;
						Overlay_Names.RelativePosition_V3 = <-42., 2.5>;
						Overlay_Names.Value = RoyalTeams[_TeamName].Members;
						Overlay_Background.Size = <85., 6. + Overlay_Names.ComputeHeight(Overlay_Names.Value)>;
					}
				}

				if (_IsMouseOut) {
					AnimMgr.Add(Background, "<quad opacity=\'0.5\' />", 350, CAnimManager::EAnimManagerEasing::ExpOut);
					AnimMgr.Add(Overlay, "<frame scale=\'0.8\' />", 500, CAnimManager::EAnimManagerEasing::ExpOut);
					Overlay.Visible = False;
				}
			}
		');

		$script->addCustomScriptLabel(ScriptLabel::ONINIT, '
			log("Script running");
			declare CMlFrame Frame_Teams <=> (Page.GetFirstChild("frame-teams") as CMlFrame);
			declare CMlLabel Label_Info <=> (Page.GetFirstChild("info-label-player-stats") as CMlLabel);
			declare CMlLabel Label_InfoText <=> (Page.GetFirstChild("info-label-text") as CMlLabel);

			declare K_RoyalTeam[Text] RoyalTeams for This;
			declare K_ServerInfo ServerInfo for This;
			declare Text[] Teams = [
				"Flamingo", "Pig", "Clown_Fish", "Fox", "Octopus", "Butterfly", "Crocodile", "Grasshopper", "Ladybug", "Macaw_Parrot",
				"Giraffe", "Bee", "Dolphin", "Peafowl", "Kangaroo", "Monkey", "Panda", "Zebra", "Rabbit", "Polar_Bear"
			];
		');

		$script->addCustomScriptLabel(ScriptLabel::LOOP, '
			Label_Info.Value = """{{{ServerInfo.NbPlayersInTeams}}} / {{{ServerInfo.NbPlayers}}}""";
			Label_InfoText.Value = " Join a team before the match starts";

			foreach (Key => Control in Frame_Teams.Controls) {
				if (Teams.existskey(Key)) {
					UpdateState((Control as CMlFrame), Teams[Key], False, False);
				}

				foreach (Event in PendingEvents) {
					if (Event.Type == CMlScriptEvent::Type::MouseOver) {
						if (TextLib::Find(Teams[Key], Event.Control.Parent.ControlId, False, False)) {
							UpdateState((Control as CMlFrame), Teams[Key], True, False);
						}
					}

					if (Event.Type == CMlScriptEvent::Type::MouseOut) {
						if (TextLib::Find(Teams[Key], Event.Control.Parent.ControlId, False, False)) {
							UpdateState((Control as CMlFrame), Teams[Key], False, True);
						}
					}

					if (Event.Type == CMlScriptEvent::Type::MouseClick) {
						if (TextLib::Find(Teams[Key], Event.Control.Parent.ControlId, False, False)) {
							TriggerPageAction("Royal.Player.JoinTeam."^Teams[Key]);
						}
					}
				}
			}
		');

		$this->maniaControl->getManialinkManager()->sendManialink($manialink, $player->login);
	}

	private function startGame(Player $player)
	{
		if (!$this->maniaControl->getAuthenticationManager()->checkRight($player, AuthenticationManager::AUTH_LEVEL_ADMIN)) {
			$this->maniaControl->getAuthenticationManager()->sendNotAllowed($player);
			return;
		}

		$players = $this->maniaControl->getPlayerManager()->getPlayers();
		foreach ($players as $player) {
			if (Teams::GetTeamIdV2($player) < 0)
				continue;

			try {
				$this->maniaControl->getClient()->triggerModeScriptEvent(self::XMLRPC_METHOD_ADDPLAYER, [
					Account::toAccountId($player->login),
					(string) Teams::GetTeamIdV2($player)
				], true);
			} catch (\Exception $e) {
				Logger::logError($e->getMessage());
			}
		}

		if (isset($this->maniaControl->getClient()->executeMulticall()[0])) {
			$this->matchStatus = self::MATCH_INPROGRESS;
			$this->maniaControl->getManialinkManager()->hideManialink(self::MLID_TEAMDISPLAY);
			$this->maniaControl->getChat()->sendSuccess("Match started");
		}
	}

	private function sendPostRequest($data, $url)
	{
		$asyncRequest = new AsyncHttpRequest($this->maniaControl, $url);
		$asyncRequest->setContentType(AsyncHttpRequest::CONTENT_TYPE_JSON);

		if (
			!empty($this->maniaControl->getSettingManager()->getSettingValue($this, self::SETTING_POST_ENDPOINT_HEADER_KEY)) &&
			!empty($this->maniaControl->getSettingManager()->getSettingValue($this, self::SETTING_POST_ENDPOINT_HEADER_VALUE))
		) {
			$asyncRequest->setHeaders([
				$this->maniaControl->getSettingManager()->getSettingValue($this, self::SETTING_POST_ENDPOINT_HEADER_KEY) . ": " .
				$this->maniaControl->getSettingManager()->getSettingValue($this, self::SETTING_POST_ENDPOINT_HEADER_VALUE)
			]);
		}

		$asyncRequest->setContent(json_encode($data));
		$asyncRequest->setCallable(function ($response, $error) {
			$response = json_decode($response);
			if ($error || !$response) {
				Logger::logError('Error while Sending data: ' . print_r($error, true));
				return;
			}

			//var_dump($response);
			$this->maniaControl->getChat()->sendSuccessToAdmins("Scores successfully submitted");
		});
		$asyncRequest->postData();
	}
}